#############################################################################
##
#W  collect.gi             zalgs package                   Tobias Moede
##
#W Contains functions for collecting in integral group rings.


######################################################################
##
#F ZAlgsCollectSameProducts( <lin> )
##  
## Returns linear combination of words, without repeated words

ZAlgsCollectSameProducts := function( lin )
local ret, i, j, new;
	
    ret := [];

    for i in lin do
        new := true;

        for j in ret do
            if i[2] = j[2] then
               j[1] := j[1]+i[1];
               new := false;
            fi;
        od;

        if new then 
            Add(ret, StructuralCopy(i));
        fi;
    od;

    return Filtered(ret, x -> x[1] <> 0);
end;


###################################################################### 
##
#F ZAlgsCollectSameProductsFlat( <lin> )
##  
## Returns linear combination of words, without repeated words

ZAlgsCollectSameProductsFlat := function( lin )
local ret, i, j, new, pos;
	
    ret := [];

    for i in [2,4..Length(lin)] do
        new := true;

        for j in [2,4..Length(ret)] do
            if lin[i] = ret[j] then
               ret[j-1] := ret[j-1]+lin[i-1];
               new := false;
            fi;
        od;

        if new then 
            Append(ret, [lin[i-1], lin[i]]);
        fi;
    od;

    pos := [];

    for i in [2,4..Length(ret)] do
        if ret[i-1] = 0 then
            Add(pos, i-1);
            Add(pos, i);
        fi;
    od;

    pos := Difference([1..Length(ret)], pos);

    return ret{pos};
end;


###################################################################### 
##
#F ZAlgsCollectProducts( <T>, <w> )
##  
## Input is a linear combination of words given in the format [ x, [ i_1, ..., i_k ] ]
## representing the word x b[i_1] ... b[i_k]
## output is a coefficient vector [a_1, ..., a_d], with no restrictions on the a_i

ZAlgsCollectProducts := function( T, w )
local todo, done, maxw, word, i, j, k, right, newfac, newprod, coeffs;

    # word has no uncollected products, do nothing
    if Length(w[2]) = 1 then
        coeffs := List([1..T.dim], x -> 0);
        coeffs[w[2][1]] := w[1];
        return coeffs;
    fi;

    # word has uncollected products
    todo := [w];
    done := [];
    maxw := T.wgs[Length(T.wgs)];

    while todo <> [] do
        # look at last word in todo
        word := StructuralCopy(todo[Length(todo)]);

        i := word[2][1];
        j := word[2][2];
        right := word[2]{[3..Length(word[2])]};

        # remove last word in todo and possibly replace by new ones
        Remove(todo);

        # trivial check for weight
        if T.wgs[i]+T.wgs[j] <= maxw then
            for k in [1..Length(T.mtab[i][j])] do
                if T.mtab[i][j][k] <> 0 then
                    newfac := word[1]*T.mtab[i][j][k];
                    newprod := Concatenation([k], right);
                    
                    if Sum(newprod, x -> T.wgs[x]) <= maxw then
                        if Length(right) = 0 then
                            Add(done, [newfac, newprod]);
                        else
                            Add(todo, [newfac, newprod]);
                        fi;
                    fi;
                fi;
            od;
        fi;
    od;

    # build coefficient vector
    coeffs := List([1..T.dim], x -> 0);
    
    for i in [1..Length(done)] do
        coeffs[done[i][2][1]] := coeffs[done[i][2][1]] + done[i][1];
    od;

    return coeffs;
end;


######################################################################
##
#F ZAlgsCollectPowers( <T>, <coeff> )
## Uses the power relations to collect an element a_1 b_1 + ... + a_d b_d
## given by its coefficient vector [a_1, ..., a_d]
ZAlgsCollectPowers := function( T, coeff )
local c, i, quot, rem;

    c := StructuralCopy(coeff);

    # go through coefficients from left to right
    for i in [1..T.dim] do
        # if corresponding power is 0 the do nothing
        if T.pows[i] = 0 then
            continue;
        fi;
 
        # coefficient too large
        if c[i] >= T.pows[i] then
            # get quotient and remainder
            quot := EuclideanQuotient(c[i], T.pows[i]);
            c[i] := EuclideanRemainder(c[i], T.pows[i]);

            # adjust coefficients for higher generators
            if ZAlgsIsZeroVec(T.ptab[i]{[i+1..T.dim]}) = false then
                c{[i+1..T.dim]} := c{[i+1..T.dim]} + quot*T.ptab[i]{[i+1..T.dim]};
            fi;
        
        # coefficient negative
        elif c[i] < 0 then
            rem := c[i] mod T.pows[i]; 
            quot := (c[i]-rem)/T.pows[i];
            c[i] := rem;
            
            # adjust coefficients for higher generators
            if ZAlgsIsZeroVec(T.ptab[i]{[i+1..T.dim]}) = false then
                c{[i+1..T.dim]} := c{[i+1..T.dim]} + quot*T.ptab[i]{[i+1..T.dim]};
            fi;
        fi;
    od;

    return c;
end;


######################################################################
##
#F ZAlgsCollectLinComb( <T>, <lin> ) 
## input is a linear combination of words given in the format [ x, [ i_1, ..., i_k ] ]
## representing the word x b[i_1] ... b[i_k]

ZAlgsCollectLinComb := function( T, lin )
local coll, i, left, right, coeffs;

	coll := ZAlgsCollectSameProducts(lin);

    # first reduce to linear combination of monomials using product relations
    for i in [1..Length(coll)] do
        coll[i] := ZAlgsCollectProducts(T, coll[i]);
    od;

    coeffs := Sum(coll);

    return ZAlgsCollectPowers(T, coeffs);
end;


######################################################################
##
#F ZAlgsMultByTable( <T>, <x>, <y> ) 
## Multiplies two elements in a table describing a consistens nilpotent presentation

ZAlgsMultByTable := function( T, x, y )
local ret, i, j;
 
    ret := ListWithIdenticalEntries(T.dim, 0);
 
    for i in [1..Length(x)] do
        for j in [1..Length(y)] do
            if x[i] <> 0 and y[j] <> 0 then
                ret := ret + x[i]*y[j]*T.mtab[i][j];
            fi;
        od;
    od;

    return ZAlgsCollectPowers(T, ret);
end;

######################################################################
##
#F ZAlgsPowerByTable( <T>, <x>, <y> ) 

ZAlgsPowerByTable := function( T, pow, coeffs )
    return ZAlgsCollectPowers(T, pow*coeffs);
end;

