#############################################################################
##
#W  nq.gi             zalgs package                   Tobias Moede
##
#W Contains all functions related to the calculation of nilpotent quotients of
#W finitely presented associative Z-algebras.


######################################################################
##
#F ZAlgsEvalAssoc( <C>, <n> )
##  
## Evaluates associativity in 'C' for the first 'n' generators.

ZAlgsEvalAssoc := function( C, n )
local R, i, j, k, left, right, res, I, maxwone;

    R := [];
    I := IdentityMat(C.dim, Integers);
    maxwone := Maximum(Positions(C.wgs, 1));

    for i in [1..maxwone] do
        for j in [1..n] do
            for k in [1..n] do
                # all other products are zero anyway
                if C.wgs[i]+C.wgs[j]+C.wgs[k] <= C.wgs[n]+1 then

                    Info(InfoZAlgs, 3, "Evaluating associativity relations for i=", i, " and j=", j, " and k=", k);
                    left := C.mtab[i][j];
                    left := ZAlgsMultByTable(C, left, I[k]);
                 
                    right := C.mtab[j][k]; 
                    right := ZAlgsMultByTable(C, I[i], right);

                    res := left-right;

                    if PositionNonZero(res) <= n  then
                        Error("[", i, ",", j, ",", k, "]-th assoc relator not in module!");
                    fi;                    

                    if ZAlgsPositionsNonzero(res) <> [] then
                        Info(InfoZAlgs, 3, "yields relator ", left-right); 
                    fi; 
                    Add(R, res);
                fi;
            od;
        od;
    od;

    return R;
end;


######################################################################
##
#F ZAlgsEvalPowerLeft( <C>, <i>, <j> )
##  
## Evaluates the left power relations in 'C' for the generators 'i' and 'j'.

ZAlgsEvalPowerLeft := function( C, i, j )
local left, right, I;

    I := IdentityMat(C.dim, Integers);

    # first type of power consistency relations

    # (e_i b_i) b_j
    left := ZAlgsPowerByTable(C, C.pows[i], I[i]);
    left := ZAlgsMultByTable(C, left, I[j]);
               
    # e_i (b_i b_j)
    right := C.mtab[i][j];
    right := ZAlgsPowerByTable(C, C.pows[i], right);
                
    return left-right;
end;


###################################################################### 
##
#F ZAlgsEvalPowerRight( <C>, <i>, <j> )
##  
## Evaluates the right power relations in 'C' for the generators 'i' and 'j'.

ZAlgsEvalPowerRight := function( C, i, j )
local left, right, I;

    I := IdentityMat(C.dim, Integers);
    
    # second type of power consistency relations

    # b_j (e_i b_i)
    left := ZAlgsPowerByTable(C, C.pows[i], I[i]);
    left := ZAlgsMultByTable(C, I[j], left);
                
    # e_i (b_j b_i)
    right := C.mtab[j][i];
    right := ZAlgsPowerByTable(C, C.pows[i], right);    
                
    return left-right;
end;


######################################################################
##
#F ZAlgsEvalPower( <C>, <n> )
##  
## Evaluates power relations in 'C' for the first 'n' generators.

ZAlgsEvalPower := function( C, n )
local R, i, j, left, right, res;

    R := [];
    
    for i in [1..n] do
        for j in [1..n] do
            if C.pows[i] > 0 then
                Info(InfoZAlgs, 3, "Evaluating power relations for i=", i, " and j=", j);

                # first type of power consistency relations
                res := ZAlgsEvalPowerLeft(C, i, j);
                Add(R, res);  
                
                if PositionNonZero(res) <= n  then
                    Error("[", i, ",", j, "]-th left power relator not in module!");
                fi;

                if ZAlgsPositionsNonzero(res) <> [] then
                     Info(InfoZAlgs, 3, "(left) yields relator: ", res);
                fi;

                # second type of power consistency relations
                res := ZAlgsEvalPowerRight(C, i, j);
                Add(R, res);  

                if PositionNonZero(res) <= n  then
                    Error("[", i, ",", j, "]-th right power relator not in module!");
                fi;

                if ZAlgsPositionsNonzero(res) <> [] then
                    Info(InfoZAlgs, 3, "(right) yields relator: ", res); 
                fi;
            fi; 
        od;
    od;

    return R;
end;


######################################################################
##
#F ZAlgsEvalImgRel( <C>, <r> )
##  
## Evaluates image relations 'r' in 'C'

ZAlgsEvalImgRel := function( C, r )
local prod, i, j, coeffs;

    coeffs := ListWithIdenticalEntries(C.dim, 0);

    for i in [2,4..Length(r)] do
        # for every summand [x, [i_1, ..., i_j]] of the relator do 
        prod := C.imgs[r[i][1]];

        for j in [2..Length(r[i])] do
            prod := ZAlgsMultByTable(C, prod, C.imgs[r[i][j]]);
        od;

        prod := ZAlgsPowerByTable(C, r[i-1], prod);
        coeffs := coeffs + prod;
    od;

    return ZAlgsCollectPowers(C, coeffs);
end;


######################################################################
##
#F ZAlgsEvalImgRel( <C>, <n> )
##  
## Evaluates image relations in 'C'

ZAlgsEvalImg := function( C, n )
local R, i, res;

    R := [];

    for i in [1..Length(C.rels)] do
        Info(InfoZAlgs, 3, "Evaluating img relation ", i, " of ", Length(C.rels), "\n");

        res := ZAlgsEvalImgRel( C, C.rels[i] );
        Add(R, res);
 
        if PositionNonZero(res) <= n  then
             Error(i, "-th image relator not in module!\n");
        fi;

        if ZAlgsPositionsNonzero(res) <> [] then
             Info(InfoZAlgs, 3, "yields relator: ", res, "\n");
        fi;
    od;

    return R;
end;


######################################################################
##
#F ZAlgsNQInit( <n>, <rels> )
##  
##  Calculates a nilpotent quotient system for P/P^2 where each generator has a definition
## 
## Input:
## n - number of generators
## rels - list of relations
##
## Output:
## A record with the following entries
## dim - the number of generators
## pows - the powers e_i
## wgs - the weights w_i
## dfs - definitions for the abstract generators b_i
## imgs - images of the generators x_i given as coefficient vectors for the b_i
## mtab - multiplication relations for the b_i
## ptab - power relations for the b_i

ZAlgsNQInit := function( n, rels )
local F, pows, i, dep, ptab, imgs, r, R, newrel, T, elim, nonelim, len;

    # convert relations of group to relations for augmentation ideal quotient
    #rels := List(Grels, x -> ZAlgsRewriteRelation(x, 8));

    R := [];
    elim := [];

    # build weight 1 relations
    for r in rels do
        newrel := List([1..n], x -> 0);
        for i in [2,4..Length(r)] do
            if Length(r[i]) = 1 then
                newrel[r[i][1]] := newrel[r[i][1]] + r[i-1];
            fi;
        od;

        Add(R, newrel);
    od;
    
    F := HermiteNormalFormIntegerMat(R);
    F := Filtered(F, x -> ZAlgsIsZeroVec(x) = false);

    # determine powers e_i and normal forms for e_i b_i
    pows := List([1..n], x -> 0);
    ptab := List([1..n], x -> ListWithIdenticalEntries(n, 0));

    for i in [1..Length(F)] do
        dep := ZAlgsDepthVector(F[i]);
        if dep <= n then
            pows[dep] := F[i][dep];
            ptab[i]{[dep+1..n]} := -F[i]{[dep+1..n]};
            
            if pows[dep] = 1 then
                Add(elim, dep); 
            fi;
        fi;
    od;

    nonelim := Difference([1..n], elim);
    len := Length(nonelim);

    # setup images
    imgs := [];
 
    for i in [1..n] do
        if i in nonelim then
            imgs[i] := ListWithIdenticalEntries(n, 0);
            imgs[i][i] := 1;
        else
            imgs[i] := StructuralCopy(ptab[i]);
        fi;
    od;

    for i in [1..n] do
        imgs[i] := imgs[i]{nonelim};
    od;

    T := rec(
        dim := len,
		pows := pows{nonelim},
        wgs := List([1..len], x -> 1),
        dfs := nonelim,
        imgs := imgs,
        rels := rels,
        mtab := List([1..len], x -> NullMat(len, len, Integers)),
		ptab := ptab{nonelim}{nonelim}
    );

    for i in [1..Length(T.imgs)] do
        T.imgs[i] := ZAlgsCollectPowers(T, T.imgs[i]);
    od;

    for i in [1..Length(T.ptab)] do
        T.ptab[i] := ZAlgsCollectPowers(T, T.ptab[i]);
    od;

    return T;
end;


######################################################################
##
#F ZAlgsNQCheckNoMul( <nqs> )
##  
## Check which elements dont get a multiplication tail

ZAlgsNQCheckNoMul := function( nqs )
local nomultail, maxwone, i, j, pos;

    nomultail := [];
    maxwone := Maximum(Positions(nqs.wgs, 1));

    for i in [1..Length(nqs.dfs)] do
        if not IsInt(nqs.dfs[i]) then
            Add(nomultail, [nqs.dfs[i][1], nqs.dfs[i][2]]);
        fi;
    od;

    for i in [maxwone+1..nqs.dim] do
        for j in [1..nqs.dim] do
            Add(nomultail, [i,j]);
        od;
    od;

    return Set(nomultail);
end;


######################################################################
## 
#F ZAlgsNQInitImg( <nqs>, <imgind>, <tgens> )
##  
## Initialize images

ZAlgsNQInitImg := function( nqs, imgind, tgens )
local i, j, imgs, ind;

    imgs := List([1..Length(nqs.imgs)], x -> ListWithIdenticalEntries(tgens, 0));

    for i in [1..Length(nqs.imgs)] do
        for j in [1..Length(nqs.imgs[i])] do
            imgs[i][j] := nqs.imgs[i][j];
        od;
    od;

    for i in [1..Length(nqs.imgs)] do
        ind := imgind[i];
        if ind <> fail then
            imgs[i][nqs.dim + ind] := 1;
        fi;
    od;

    return imgs;
end;


######################################################################
## 
#F ZAlgsNQInitPow( <nqs>, <powind>, <tgens> )
##  
## Initialize powers 

ZAlgsNQInitPow := function( nqs, powind, tgens )
local i, j, ptab, ind;

    ptab := List([1..tgens], x -> ListWithIdenticalEntries(tgens, 0));

    for i in [1..nqs.dim] do
        for j in [1..nqs.dim] do
            ptab[i][j] := nqs.ptab[i][j];
        od;
    od;

    # attach tail to relation e_i b_i
    for i in [1..nqs.dim] do
        ind := powind[i];
        if ind <> fail then
            ptab[i][nqs.dim + ind] := 1;
        fi;
    od;
 
    return ptab;
end;


######################################################################
## 
#F ZAlgsNQInitMul( <nqs>, <mulind>, <tgens> )
##  
## Initialize multiplication table

ZAlgsNQInitMul := function( nqs, mulind, tgens )
local maxwone, i, j, k, mtab, ind;

    mtab := List([1..tgens], x -> NullMat(tgens, tgens, Integers));
    maxwone := Maximum(Positions(nqs.wgs, 1));
 
    # init relations b_i b_j for b_i of weight 1
    for i in [1..maxwone] do
        for j in [1..nqs.dim] do
            for k in [1..nqs.dim]  do
                mtab[i][j][k] := nqs.mtab[i][j][k];
            od;

            ind := mulind[i][j];
            if ind <> fail then
                mtab[i][j][nqs.dim + ind] := 1;
            fi;
        od;
    od;

    return mtab;
end;


######################################################################
##
#F ZAlgsRewriteIteratedWOne( <nqs>, <prod> )
##  
## Rewrites a word as a product of weight one elements using the definitions

ZAlgsRewriteIteratedWOne := function( nqs, prod )
local maxwone, ret, pos, new, i;

    ret := [];
    maxwone := Maximum(Positions(nqs.wgs, 1));

    # positions of non-weight one elements
    for i in [1..Length(prod)] do
        if nqs.wgs[prod[i]] = 1 then
            Add(ret, prod[i]);
        else
            Append(ret, nqs.dfs[prod[i]]);
        fi;
    od;

    return ret;
end;


######################################################################
##
#F ZAlgsNQCompleteMul( <C> )
##
## Completes a multiplication table using definitions 

ZAlgsNQCompleteMul := function( C )
local maxwone, maxcomp, i, j, k, prod, coeffs, I;

    maxwone := Maximum(Positions(C.wgs, 1));
    maxcomp := Minimum(Positions(C.wgs, 0))-1;
    I := IdentityMat(C.dim, Integers);

    # complete table using definitions
    for i in [maxwone+1..maxcomp] do
        for j in [1..maxcomp] do
            if C.wgs[i]+C.wgs[j] <= C.wgs[maxcomp]+1 then
                # which product to calculate
                prod := ZAlgsRewriteIteratedWOne(C, [i,j]);   
            
                # work from right to left and calculate iterated product
                coeffs := ZAlgsMultByTable(C, I[prod[Length(prod)-1]], I[prod[Length(prod)]]);

                # check if already trivial
                if ZAlgsIsZeroVec(coeffs) then
                    C.mtab[i][j] := StructuralCopy(coeffs);
                    continue;
                fi;

                for k in Reversed([1..Length(prod)-2]) do
                     coeffs := ZAlgsMultByTable(C, I[prod[k]], coeffs);
                     if ZAlgsIsZeroVec(coeffs) then
                        break;
                     fi;
                od; 

                C.mtab[i][j] := StructuralCopy(coeffs);
            fi;
        od;
    od;
end;



######################################################################
## 
#F ZAlgsNQExtend( <nqs>, <prod> )
##
## Calculates a nilpotent quotient system for P/P^{c+1} where each generator has a definition
## 
## Input:
## A NQ system for the class c quotient as 
##
## Output:
## A NQ system for the class c+1 quotient

ZAlgsNQExtend := function( nqs, rels )
local maxwone, num_tails, imgind, powind, i, mulind, j, num_imgtails, num_powtails, num_multails, imgs, ptab, mtab, pows, wgs, dim, dfs,
      R, dep, replacer, elim, class, C, t, k, pos, fac, nonelim, I, prods, Rimg, ncol;

    #rels := List(Grels, x -> ZAlgsRewriteRelation(x, c+1));
    maxwone := Maximum(Positions(nqs.wgs, 1));
    class := nqs.wgs[Length(nqs.wgs)];

    # create index of tails
    
    # create img index (not done yet)
    num_tails := 0;
    imgind := [];

    for i in [1..Length(nqs.imgs)] do
        if not i in nqs.dfs then
            num_tails := num_tails+1;
            Add(imgind, num_tails);
        else
            Add(imgind, fail);
        fi;
    od;

    num_imgtails := num_tails;

    # create power index
    powind := [];

    for i in [1..nqs.dim] do
        if nqs.pows[i] <> 0 then
            num_tails := num_tails + 1;
            Add(powind, num_tails);
        else
            Add(powind, fail); 
        fi;
    od;

    num_powtails := num_tails - num_imgtails;

    # create mult index (last with correct weight)
    mulind := List([1..nqs.dim], x -> List([1..nqs.dim], y -> fail));

    for i in [1..maxwone] do
        for j in [1..nqs.dim] do
            if (not [i,j] in nqs.dfs) and (nqs.wgs[i]+nqs.wgs[j] < class+1) then
                num_tails := num_tails + 1; 
                mulind[i][j] := num_tails;  
            fi;
        od;
    od;

    for i in [1..maxwone] do
        for j in [1..nqs.dim] do
            if (not [i,j] in nqs.dfs) and (nqs.wgs[i]+nqs.wgs[j] = class+1)  then
                num_tails := num_tails + 1;  
                mulind[i][j] := num_tails;    
            fi;
        od;
    od;

    num_multails := num_tails - num_imgtails - num_powtails;

    Info(InfoZAlgs, 3, "introduced ", num_imgtails, " image tails\n");
    Info(InfoZAlgs, 3, "introduced ", num_powtails, " power tails\n");
    Info(InfoZAlgs, 3, "introduced ", num_multails, " multiplication tails\n");

    # create larger table with tails
    #####
    # initialize new imgs
    imgs := ZAlgsNQInitImg(nqs, imgind, nqs.dim + num_tails);

    #####
    # initialize new ptab
    ptab := ZAlgsNQInitPow(nqs, powind, nqs.dim + num_tails);
 
    #####
    # initalize new mtab
    mtab := ZAlgsNQInitMul(nqs, mulind, nqs.dim + num_tails);

    #####
    # initialize new dim, pows and weights
    pows := Concatenation(nqs.pows, ListWithIdenticalEntries(num_tails, 0));

    # new weights are unknown!
    wgs := Concatenation(nqs.wgs, ListWithIdenticalEntries(num_tails, 0));
    dim := nqs.dim + num_tails;

    # setup new definitions
    dfs := List([1..dim], x -> 0);

    # copy existing definitions
    for i in [1..nqs.dim] do
        dfs[i] := nqs.dfs[i];
    od;

    # build new definitions 
    for i in [1..num_powtails] do
        dfs[nqs.dim+num_imgtails+i] := [i];
    od;

    for i in [1..num_multails] do
        dfs[nqs.dim+num_imgtails+num_powtails+i] := ZAlgsFindInTwoDim(mulind, num_imgtails+num_powtails+i);
    od;

    # Evaluate consistency relations in this larger table
    C := rec( dfs := dfs,
              dim := dim,
              imgs := imgs,
              mtab := mtab, 
              pows := pows,
              ptab := ptab, 
              rels := rels,
              wgs := wgs, powind := powind, mulind := mulind );

    # Complete mtab
    ZAlgsNQCompleteMul(C);

    R := [];
    
    Info(InfoZAlgs, 1, "Evaluating associativity consistency"); 
    t := Runtime();
    Append(R, ZAlgsEvalAssoc(C, nqs.dim));
    Info(InfoZAlgs, 2, "took ", Runtime()-t, "ms");

    Info(InfoZAlgs, 1, "Evaluating power consistency"); 
    t := Runtime();
    Append(R, ZAlgsEvalPower(C, nqs.dim));
    Info(InfoZAlgs, 2, "took ", Runtime()-t, "ms");

    Info(InfoZAlgs, 1, "Evaluating image consistency"); 
    t := Runtime();
    Append(R, ZAlgsEvalImg(C, nqs.dim));
    Info(InfoZAlgs, 2, "took ", Runtime()-t, "ms");

    # calculate hermite normal form of relations and filter only non-zero rows
    R := HermiteNormalFormIntegerMat(R);  
   
    # these are image tails
    Rimg := Filtered(R, x -> PositionNonZero(x) <= nqs.dim + num_imgtails);

    # these are remaining relators
    R := Filtered(R, x -> ZAlgsIsZeroVec(x) = false);
    R := Filtered(R, x -> PositionNonZero(x) > nqs.dim + num_imgtails);

    # powers can be read of hermite normal form and ptab can be adjusted
    for i in [1..Length(R)] do
        dep := ZAlgsDepthVector(R[i]);
        pows[dep] := R[i][dep];
        ptab[dep] := Concatenation( ListWithIdenticalEntries(dep, 0), -R[i]{[dep+1..Length(R[i])]});
    od;

    # Replace power relations and eliminate
    elim := [];
    replacer := [];
 
    # eliminate all image tails and rewrite images
    for i in [1..num_imgtails] do
        Add(elim, nqs.dim+i);
    od;

    for i in [1..Length(Rimg)] do
        dep := ZAlgsDepthVector(Rimg[i]); 
        pos := Position(imgind, dep-nqs.dim);
        imgs[pos]{[dep+1..Length(Rimg[i])]} := -Rimg[i]{[dep+1..Length(Rimg[i])]};
        imgs[pos] := ZAlgsCollectPowers(C, imgs[pos]);
    od;

    for i in [1..Length(R)] do
        dep := ZAlgsDepthVector(R[i]);
        if R[i][dep] = 1 then
            # store which vectors can be rewritten
            if not ZAlgsIsZeroVec(R[i]{[dep+1..Length(R[i])]}) then
                # rewrite [0, ..., 1, 0, ..., 0] as [0, ..., 0, -*, ..., -*]
                Add(replacer, [dep, Concatenation(ListWithIdenticalEntries(dep, 0), -R[i]{[dep+1..Length(R[i])]})]);
            fi;

            # eliminate generator
            Add(elim, dep);
        fi;
    od;

    Info(InfoZAlgs, 3, "Replace using the following rules:\n", replacer);
    Info(InfoZAlgs, 3, "Eliminate the following generators:\n", elim);
    t := Runtime();

    # replace and collect
    for i in [1..Length(mtab)] do
        for j in [1..Length(mtab)] do
            ncol := false;

            for k in [1..Length(replacer)] do
                pos := replacer[k][1];
                fac := mtab[i][j][pos];

                if fac <> 0 then
                    ncol := true;
                    mtab[i][j][pos] := 0;
                    mtab[i][j] := mtab[i][j] + fac*replacer[k][2];
                fi;
            od;

            if ncol then
                 mtab[i][j] := ZAlgsCollectPowers(C, mtab[i][j]);
            fi;
        od;
    od;

    for i in [1..Length(ptab)] do
        ncol := false;

        for k in [1..Length(replacer)] do
            pos := replacer[k][1];
            fac := ptab[i][pos];
              
            if fac <> 0 then
                ncol := true;
                ptab[i][pos] := 0;
                ptab[i] := ptab[i] + fac*replacer[k][2];
            fi;
        od;

        if ncol then
            ptab[i] := ZAlgsCollectPowers(C, ptab[i]);
        fi;
    od;

    Info(InfoZAlgs, 2, "Replacing took: ", Runtime()-t, "ms");
    t := Runtime();

    nonelim := Difference([1..dim], elim);
  
    # cut everything
    dfs := dfs{nonelim};
    dim := Length(nonelim);
    mtab := mtab{nonelim}{nonelim}{nonelim};
    pows := pows{nonelim};        
    ptab := ptab{nonelim}{nonelim};
    wgs := wgs{nonelim};
    
    # adjust weights
    for i in [1..Length(wgs)] do
        if wgs[i] = 0 then
            wgs[i] := class+1;
        fi;
    od;

    for i in [1..Length(imgs)] do
        imgs[i] := imgs[i]{nonelim};
    od;

    Info(InfoZAlgs, 2, "Cutting took: ", Runtime()-t, "ms");

    # return a nilpotent quotient system for A_{c+1}
    return rec( dfs := dfs,
                dim := dim,
                imgs := imgs,
                mtab := mtab, 
                pows := pows,
                ptab := ptab, 
                rels := rels,
                wgs := wgs );
end;



######################################################################
## 
#F NilpotentQuotientFpZAlgebra( <A>, <c> )
##
## Build the class-'c' quotient of a finitely presented associative Z-algebra

InstallMethod( 
        NilpotentQuotientFpZAlgebra,
        "for finitely presented associative Z-algebras",
        [ IsAlgebra, IsInt ],
        function( A, c )

local numgens, grels, quot, i;

    numgens := Size(GeneratorsOfAlgebra(A));
    grels := ZAlgsBuildGrelsFpAlgebra(A);
    quot := ZAlgsNQInit(numgens, grels);

    for i in [2..c] do
        quot := ZAlgsNQExtend(quot, grels);
    od;

    return quot;
end );


######################################################################
## 
#F AugmentationQuotientFpGroup( <G>, <c>, <print> )
##
## Build the class-'c' quotient of the augmentation ideal of a finitely
## presented group 'G'. If the optional argument 'print' is set to true
## the structure of the augmentation quotients I^i/I^(i+1) is printed
## during computation.

InstallMethod( 
        AugmentationQuotientFpGroup,
        "for finitely presented groups",
        [ IsFpGroup, IsInt, IsBool ],
        function( G, c, print... )

local numgens, grels, quot, i;

    numgens := Size(GeneratorsOfGroup(G));
    grels := ZAlgsBuildGrelsFpGroup(G, c+1);
    quot := ZAlgsNQInit(numgens, grels);

    if Length(print) > 0 and print[1] = true then
        ZAlgsPrintAugmentationQuotient(quot, 1);
    fi;

    for i in [2..c] do
        quot := ZAlgsNQExtend(quot, grels);

        if Length(print) > 0 and print[1] = true then
            ZAlgsPrintAugmentationQuotient(quot, i);
        fi;
    od;

    return quot;
end );


######################################################################
## 
#F ZAlgsAugmentationQuotientPcpGroup( <G>, <c>, <print> )
##
## Build the class-'c' quotient of the augmentation ideal of a 
## pcp-group 'G'. If the optional argument 'print' is set to true
## the structure of the augmentation quotients I^i/I^(i+1) is printed
## during computation.

InstallMethod( 
        AugmentationQuotientPcpGroup,
        "for pcp-groups",
        [ IsPcpGroup, IsInt, IsBool ],
        function( G, c, print... )

local numgens, grels, quot, i;

    numgens := Size(GeneratorsOfGroup(G));
    grels := ZAlgsBuildRelsPcpGroup(G, c+1);
    quot := ZAlgsNQInit(numgens, grels);

    if Length(print) > 0 and print[1] = true then
        ZAlgsPrintAugmentationQuotient(quot, 1);
    fi;

    for i in [2..c] do
        quot := ZAlgsNQExtend(quot, grels);

        if Length(print) > 0 and print[1] = true then
            ZAlgsPrintAugmentationQuotient(quot, i);
        fi;
    od;

    return quot;
end );
