#############################################################################
##
#W  rewriterel.gi             zalgs package                   Tobias Moede
##
#W Contains functions for converting group relations to relations in quotients
#W of the associated integral group rings.


######################################################################
##
#F ZAlgsMultiplySums( <s>, <t> )

ZAlgsMultiplySums := function( s, t )
local ret, i, j;
    
    ret := [];

    for i in [1..Length(s)] do
        for j in [1..Length(t)] do
            if s[i][1] <> 0 and t[j][1] <> 0 then
                Add(ret, [s[i][1]*t[j][1], Concatenation(s[i][2], t[j][2])]);
            fi;
        od;
    od;

    return ret;
end;


######################################################################
##
#F ZAlgsRewriteInverse( <j>, <n> )
##  
## Rewrites (g_j^-1 - 1) in I/I^(n+1)

ZAlgsRewriteInverse := function( j, n )
local i, ret;

    ret := [];

    for i in [1..n] do
        Add(ret, [(-1)^i, ListWithIdenticalEntries(i, j)]);
    od;
  
    return ret;
end;


######################################################################
##
#F ZAlgsRewriteRelation( <rel>, <n> )
##  
## Rewrites a relation [i_1, e_1, ... , i_s, e_s] in I/I^(n+1)

ZAlgsRewriteRelation := function( rel, n )
local ret, s, i, j, t, prod;

    s := Length(rel)/2;
    ret := [];

    # loop over all sizes of subsets
    for i in [1..s] do
        # products for all i-element subsets of {1,...,s}
        for t in IteratorOfCombinations([1..s], i) do
            # e_1 = 1
            if rel[2*t[1]] = 1 then
                prod := [ [1, [rel[2*t[1]-1]]] ];
            # e_1 = -1
            else
                prod := ZAlgsRewriteInverse(rel[2*t[1]-1], n);
            fi;

            # extend product
            for j in [2..i] do
                if rel[2*t[j]] = 1 then
                    prod := ZAlgsMultiplySums(prod, [ [1, [rel[2*t[j]-1]]] ]);
                else
                    prod := ZAlgsMultiplySums(prod, ZAlgsRewriteInverse(rel[2*t[j]-1], n));
                fi;
            od;

            Append(ret, prod);
        od;
    od;

    ret := Filtered(ret, x -> Length(x[2]) < n+1);

    return Concatenation(ZAlgsCollectSameProducts(ret));
end;


######################################################################
##
#F ZAlgsNegateRelation( <rel> )
##  
## Multiplies all factors in 'rel' by -1

ZAlgsNegateRelation := function( rel )
local i;

    for i in [1,3..Length(rel)-1] do
        rel[i] := -rel[i];
    od;
end;


######################################################################
##
#F ZAlgsBuildRelsPcpGroup( <G>, <n> )
##  
## Builds relations for the I/I^n+1 for a given pcp-group 'G'

ZAlgsBuildRelsPcpGroup := function ( G, n )
local nil, pcp, gens, ords, i, r, g, j, h, c, lhs, rhs, rels, k, l, exp;
    
    nil := IsNilpotent(G);
    pcp := Pcp(G);
    gens := GeneratorsOfPcp(pcp);
    ords := RelativeOrdersOfPcp(pcp);
    rels := [];

    for i in [1..Length(gens)] do
        if ords[i] > 0 then
            r := ords[i];
            g := gens[i];

            lhs := [];
            rhs := [];

            # lhs of g_i^r_i = ...
            for k in [1..r] do
                Add(lhs, i);
                Add(lhs, 1);
            od; 

            lhs := ZAlgsRewriteRelation(lhs, n);
     
            # rhs of g_i^r_i = ... 
            exp := ExponentsByPcp(pcp, g^r);
            for k in [1..Length(exp)] do
                for l in [1..exp[k]] do
                    Add(rhs, k);
                    Add(rhs, 1);
                od;
            od;             

            rhs := ZAlgsRewriteRelation(rhs, n);
            ZAlgsNegateRelation(rhs);
            Add(rels, ZAlgsCollectSameProductsFlat(Concatenation(lhs, rhs)));
        fi;
    od;

    for i in [1..Length(gens)] do
        for j in [1..i-1] do
            g := gens[i];
            h := gens[j];
            c := g^h;

            # just commuting
            if c=g then
                Add(rels, [ 1, [i,j], -1, [j,i] ]);
            else
                lhs := [];
                rhs := [j,1];

                # lhs of g_i g_j = g_j ...
                lhs := ZAlgsRewriteRelation([i,1,j,1], n);

                # rhs of g_i g_j = g_j ...
                exp := ExponentsByPcp(pcp, c);

                for k in [1..Length(exp)] do
                    for l in [1..AbsInt(exp[k])] do
                        Add(rhs, k);
                        if exp[k] > 0 then
                            Add(rhs, 1);
                        else
                            Add(rhs, -1);
                        fi;
                    od;
                od;   

                rhs := ZAlgsRewriteRelation(rhs, n);
                ZAlgsNegateRelation(rhs);
                Add(rels, ZAlgsCollectSameProductsFlat(Concatenation(lhs, rhs))); 
            fi;

            if ords[j] = 0 and not nil then
                c := g^(h^-1);

                # just commuting
                if c=g then
                    Add(rels, [ 1, [j,i], -1, [i,j] ]);
                else
                    lhs := [];
                    rhs := [j,-1];

                    # lhs of g_i g_j^-1 = g_j^-1 ...
                    lhs := ZAlgsRewriteRelation([i,1,j,-1], n);

                    # rhs of g_i g_j = g_j ...
                    exp := ExponentsByPcp(pcp, c);

                    for k in [1..Length(exp)] do
                        for l in [1..AbsInt(exp[k])] do
                            Add(rhs, k);
                            if exp[k] > 0 then
                                Add(rhs, 1);
                            else 
                                Add(rhs, -1);
                            fi;
                        od;
                    od;   

                    rhs := ZAlgsRewriteRelation(rhs, n);
                    ZAlgsNegateRelation(rhs);
                    Add(rels, ZAlgsCollectSameProductsFlat(Concatenation(lhs, rhs))); 
                fi;
            fi;
        od;
    od;

    return rels;
end;


###################################################################### 
##
#F ZAlgsBuildRelsFpGroup( <G>, <n> )
##  
## Builds relations for the I/I^n+1 for a given fp-group 'G'

ZAlgsBuildGrelsFpGroup := function( G, n )
local grels, r, ext, rel, i, j;

    grels := [];

    for r in RelatorsOfFpGroup(G) do
        rel := [];
        ext := ExtRepOfObj(r); 

        for i in [2,4..Length(ext)] do
            if ext[i] > 0 then
                for j in [1..ext[i]] do
                    Add(rel, ext[i-1]);
                    Add(rel, 1);
                od;
            else
                for j in [1..AbsInt(ext[i])] do
                    Add(rel, ext[i-1]);
                    Add(rel, -1);
                od;
            fi;
        od;

        Add(grels, rel);  
    od;

    grels := List(grels, x -> ZAlgsRewriteRelation(x, n));

    return grels;
end;


###################################################################### 
##
#F ZAlgsBuildRelsFpZAlgebra( <A> )
##  
## Builds relations for a given fp-group algebra 'A' over Z

ZAlgsBuildGrelsFpAlgebra := function( A )
local grels, r, ext, rel, i, j, k;

    grels := [];

    for r in RelatorsOfFpAlgebra(A) do
        rel := [];
        ext := ExtRepOfObj(r)[2]; 

        for i in [2,4..Length(ext)] do
            Add(rel, ext[i]);
            Add(rel, []);
            
            for j in [2,4..Length(ext[i-1])] do
                for k in [1..ext[i-1][j]] do
                    Add(rel[i], ext[i-1][j-1]);
                od;
            od;
        od;

        Add(grels, rel);  
    od;

    return grels;
end;
